/* This class is a thread that is created by the PCCE server to handle a connection. 
   One thread per connection will be created to process requests from clients.
*/

import pccejava.*;
import pccejava.NetComm;
import pccejava.SSLComm;
import java.net.*;
import java.util.*;
import java.io.*;
import javax.net.ssl.*;

class ClientThread extends Thread {
    // private variables for network communication and user information.
    private PCCEServer server;
    // network communication object used by this thread, which
    // communicates with a remote client 	   
    //private NetComm myComm = null;
    private SSLComm myComm = null;
    private FileHandler fileHandler = null;	
    // variables to store user information
    private String usrid;
    // key into hashtable of user objects
    private String sessionId;	
    // variables for user information	
    private String tool;   // tool user is launching	
    private String passwd;
    private String senderHost;
    private String realname;
    private String emailAddr = "NONE";
    private String company = "NONE";
    private String title = "NONE";	
    private String groups = "NONE";  
    private String nickChange = null;
    private User thisUser; 
    private boolean DEBUG = true;
    private boolean authenticated = false;

    // Count number of heartbeats sent to client.
    private int numPings = 0;

    ClientThread(SSLSocket s, PCCEServer caller, FileHandler fh, 
    		String f1, String f2, String p, Vector v) 
    {
	super("PCCEServer");
	myComm = new SSLComm(null, 7310, f1, f2, p);
	myComm.setClientSock(s);	
	myComm.setPeerDN(v);	
	// Reads should timeout to check connection status. 
	try {
	    myComm.setSocketTimeout(120000);
	} catch(SocketException e) { 
	    System.out.println(new Date() + 
		": ClientThread constructor--SocketException: "+e.getMessage());
	}
	server = caller;
	fileHandler = fh;
	if (v != null)
	    authenticated = true; 	
    } 

    public void run() {
	/* While there's a connection, read data, process the message,
           and send a reply.
        */
	while (myComm.isConnected() == true) 
        {
	     /* If we get a null msg, client closed the connection, so break
	        out of while loop, close the connection, and return (which
		should kill this thread).
	     */
             String request = null;
             String reply = null;
		
	     try {
		/*
		Runtime rt = Runtime.getRuntime();
		long total = rt.totalMemory();
		long free = rt.freeMemory();
		System.out.println("Total mem: " + total);
		System.out.println("Free mem: " + free);
		System.out.println("Mem Used: " + (total - free));
		System.out.println("Max mem: " + rt.maxMemory());
		rt = null;
	 	*/

	     	request = myComm.getMessage();
		if (numPings > 0)
		    numPings--;
	     	if (DEBUG)	
	            System.out.println("ClientThread(" + usrid +  "): " + 
			new Date() + ": got request: " + request + 
			", calling handleRequest");

		 reply = handleRequest(request);
		 if (DEBUG)
	            System.out.println("ClientThread(" + usrid +  "): " + 
			new Date() + ": handleRequest ret'd reply: " + reply); 
		    
		 if (reply != null)
		 {
		     if (reply.equals("disconnect"))
		     {
		         System.out.println("ClientThread(" + usrid + "): " + 
			    new Date() + ": breaking out of run() for disconnect");
			 break;
		     }
		     else
		     {
	      	         myComm.writeMessage(reply);
			 if (DEBUG)
		             System.out.println("ClientThread(" + usrid + 
			        "): " + new Date() + ": sent reply: " + reply +
			        " , myComm: " + myComm + ", sslsock: " +
			        myComm.getClientSSLSock());
		     }
		 }
	    } catch(InterruptedIOException e) { 
	   	 System.out.println("ClientThread(" + usrid + 
		     ")::getMessage() timed out");
		 /*
		 if (numPings > 2)
		     System.out.println("***** " + usrid + " NOT RESPONDING");
		 myComm.writeMessage("PING " + usrid);
		 numPings++;
		 */
	         continue; 
	    } catch(EOFException e) {
	        System.out.println("ClientThread(" + usrid + 
		    ")::handleRequest EOFException: " + e.getMessage());  
	 	e.printStackTrace();
	        break;
	    } catch(IOException e) {
	        System.out.println("ClientThread(" + usrid +
	  	    ")::handleRequest IOException: " + e.getMessage());  
	 	e.printStackTrace();
	        break;
	    } catch(Exception e) {
	        System.out.println("ClientThread(" + usrid +
	            ")::handleRequest exception: " + e.getMessage()); 
	 	e.printStackTrace();
	        break;
	    }
	} 
	myComm.cleanup();
	myComm = null;
	server.deleteConnection(this);
    }

    private String handleRequest(String request) 
    {
	String reply = "NONE";

        // Extract the message type as the first field.
        StringTokenizer t = new StringTokenizer(request);
        String msgType = t.nextToken().toLowerCase();
	    
        if (DEBUG)
           System.out.println("ClientThread(" + usrid + 
		")::" + new Date() + ": handleRequest--msgType: " + msgType);
        if (t.hasMoreTokens()) 
        {
	   int i = request.indexOf(" ");
	   if (i < 1)
	   	return(msgType + " BADREQUEST");
           if (msgType.equals("login"))
           {
	       reply= doLogin(request.substring(i+1));
               if (DEBUG)
                   System.out.println("ClientThread(" + usrid +
                        "): " + new Date() + ": doLogin returned reply: " + reply);
           }
           else if (msgType.equals("register"))
           {
               reply = doRegister(request.substring(i+1));
               if (DEBUG)
                    System.out.println("ClientThread(" + usrid +
                        "): " + new Date()+ ": doRegister returned reply: "+reply);
           }
           else if (msgType.startsWith("changereg"))
           {
               reply = changeRegInfo(request.substring(i+1));
               if (DEBUG)
                    System.out.println("ClientThread(" + usrid +
                       "): "+new Date()+": changeRegInfo returned reply: "+reply);
           }
           else if (msgType.toLowerCase().startsWith("newreg"))
           {
               reply = newRegistration(request.substring(i+1));
               if (DEBUG)
                    System.out.println("ClientThread(" + usrid +
                       "): " + new Date() + ": newRegistration returned reply: " 
		       + reply);
           }
           else if (msgType.equals("logout"))
               server.doLogout(request.substring(i+1));
           else if (msgType.equals("userinfo"))
               reply = server.getUserInfo(request.substring(i+1));
           else if (msgType.equals("isLoggedIn"))
           {
               if (server.sessionExists(request.substring(i+1)) == true)
                    reply = "true";
               else
                    reply = "false";
           }
           else if (msgType.equals("getnick"))
                reply = getNickFromDatabase(request.substring(i+1));
           else if (msgType.equals("identify"))
                reply = identifyByName(request.substring(i+1));
           else if (msgType.equals("join"))
                reply = forwardJoinReq(request.substring(i+1));
           else if (msgType.equals("ping"))
                reply = forwardPingReq(request.substring(i+1));
	   else if (msgType.equals("online"))
		reply = server.getOnlineUsers(request.substring(i+1));
	   else if (msgType.equals("status"))
		reply = doStatus(request.substring(i+1));
	   else if (msgType.equals("location"))
		reply = doLocation(request.substring(i+1));
	   else if (msgType.equals("newvenue"))
		reply = newVenue(request.substring(i+1));
	   else if (msgType.equals("removevenue"))
		server.deleteChannel(request.substring(i+1));
	   else if (msgType.equals("door"))
		reply = doDoor(request.substring(i+1));
	   else if (msgType.equals("putnote"))
		reply = server.putNote(usrid, request.substring(i+1));
	   else if (msgType.equals("get"))
		reply = doGet(request.substring(i+1));
	   else if (msgType.equals("put"))
		reply = doPut(request.substring(i+1));
	   else if (msgType.equals("delete") || msgType.equals("remove"))
		reply = doDelete(request.substring(i+1));
	   else if (msgType.equals("ircd"))
	   {
		reply = server.doIRCServer(request.substring(i+1));
		myComm.sendMessage(reply);
		return null;
	   }
           else
                reply = "Unknown message type";
        }
        else 
        {
           if (msgType.equals("pchan") || msgType.equals("tchan"))
                reply = server.getChannels(msgType); 
	   else if (msgType.equals("mcastaddrs"))
		reply = server.getMulticastAddrs();
	   else if (msgType.equals("allusers"))
	        reply = getEntries("user", "allusers");
	   else if (msgType.toLowerCase().equals("ping"))
		reply = "pong"; 
	   else if (msgType.toLowerCase().equals("disconnect"))	
		return "disconnect";
        }
	return reply;
    }

    /* Given a string in the form: password [Real Name], return the nick from	
       the database.
    */
    public String getNickFromDatabase(String str)
    {
	String passwd = null;
	String name = null;
	String[] nick = null;
	int i = str.indexOf("[");
	if (i > 1)
        {
	    passwd = str.substring(0, (i-1));
	    name = str.substring(i); 	 
	}
	if (name != null && name.length() > 2) 		
	    nick = fileHandler.findAllByKeys("user", name);
	if (nick == null)
	    return null;
	StringTokenizer t;
	for (i=0; i<nick.length; i++) 
	{
	    t = new StringTokenizer(nick[i]);
	    if (t.countTokens() > 2)
	    {
	        name = t.nextToken();
	        if (authenticated || t.nextToken().equals(passwd))
		    return name;
	    }	
	}
	return null;
    }

    /* This method finds the nick for the DN in the client's X.509 certificate.
       It assumes an X.509 cert was successfully given by the client when connecting. 	
    */
    public String identifyByName(String msg)
    {
	/* msg form (backward compatible):
	    	<tool> <senderHost> <passwd> [Full Name]
	   The last 2 fields are ignored.	
	*/
	StringTokenizer t = new StringTokenizer(msg);
	if (t.countTokens() < 2)
	    return "identify BADREQUEST";
	tool = t.nextToken();
	senderHost = t.nextToken();
	usrid = verifyCert();
	if (usrid == null)
	    return "identify CERTNOTKNOWN";
	String[] usr = fileHandler.findAllByKeys("user", usrid);
	if (usr == null || usr.length == 0)
	    return "identify NOTREGISTERED";
	// At this point we have a database entry.
	t = new StringTokenizer(usr[0]);
	if (t.countTokens() < 3)
	    return "identify NOTREGISTERED";
	usrid = t.nextToken();
	passwd = t.nextToken();
	int i = usr[0].indexOf("[");
	if (i < 0)
	    return "identify NOTREGISTERED";
	usrid = server.getUserid(usrid);
	sessionId = usrid + senderHost;
	msg = usr[0].substring(i);
	parseUserInfo(msg);
	server.addSessionId(this, sessionId);
	server.addUser(sessionId, createUser(), tool);
	return "identify SUCCESS " + sessionId + " " + usrid + " " + getState();
    }

    private User createUser()
    {
	thisUser= new User(sessionId);
	thisUser.setUserInfo(usrid, passwd, realname, emailAddr, company, title,
	    senderHost, tool);
	thisUser.setAuthenticated(authenticated);
	StringTokenizer t = new StringTokenizer(groups);
	while (t.hasMoreTokens())
	    thisUser.addGroup(t.nextToken());
	return thisUser;
    }	

    // Get the nick from the database that is associated with the DN in the cert.
    private String verifyCert()
    {
	String nick = null;
	Vector v = myComm.getPeerDN();
	if (v != null)
	{
	    // To guarantee uniqueness concatenate DN of user w/ DNs of CAs.
	    StringBuffer buff = new StringBuffer();
	    for (int i=0; i<v.size(); i++)
	    {
	 	buff.append((String)v.elementAt(i));
		buff.append(", ");
	    }
	    String dn = buff.toString();
	    dn.trim();
	    dn = dn.substring(0, dn.length());
	    int i = dn.lastIndexOf(",");
	    if (i > 0)
		dn = dn.substring(0, i);
	    String[] dbentries = fileHandler.findAllByKeys("dn", dn);
	    
	    if (dbentries != null && dbentries.length > 0)
	    {
	 	i = dbentries[0].indexOf(" ");
		if (i > 0)
		    nick = new String(dbentries[0].substring(0,i));	
		authenticated = true;
	    }
	}
	return nick;	
    } 

    public String doLogin(String msg) 
    {
	if (DEBUG)
	    System.out.println("ClientThread(" + usrid + "): doLogin()--msg: " + 
	      msg);
 	// Parse header to set tool, usrid, sessionId, senderHost, and passwd.
	msg = parseHeader(msg);
	if (msg != null && msg.equals("BADHEADER"))
	    return "login BADREQUEST";	

        /* If user is logging in with the usrid used in the current nickname, do 
	    nothing: get the current nickname and return ISLOGGEDIN if the current
	    nickname starts with this usrid.  
        */
	String nick = null; 
	   
	/* If usrid equals "guest", accept it w/o searching the database. (Akenti 
	   authorization will be used in the future.)  If usrid is not "guest", do the 
           following:   Search the database for the usrid. If usrid is found, compare 
	   passwords.  If passwords match, add a User object to the hashtable
 	   and return SUCCESS; otherwise return BADPASSWD.  If usrid isn't in 
	   the database, return NOTREGISTERED. 
        */	   	

	if (usrid.equals("guest"))
	{
	    /* This is the first login as 'guest'. It may be for new session or it may
	       be a new login for an existing session.  In either case, get a unique
	       nickname.  Add a User object if it is a new session. For an existing 
	       session, change the login information in the existing User object.
            */
	    if (parseUserInfo(msg) == 0)
	        return("login BADREQUEST");
	    if (realname.equals("") || realname.equals("NONE") || realname.equals(" "))
		return("login BADREQUEST");

	    usrid = server.getUserid(usrid);

	    /* Extract the realname, email address, company, and title and call 
	       PCCEServer to add this user.
             */
	    server.addSessionId(this, sessionId);
	    if (server.getNick(sessionId) == null)
	        server.addUser(sessionId, createUser(), tool);
	    else
	        server.changeLoginInfo(sessionId, usrid, usrid);
	    return("login SUCCESS " + sessionId + " " + usrid + " " + getState());
	 }
	 String reply = fileHandler.findUser(usrid); 
	 if (reply == null)
	     reply = "login NOTREGISTERED " + sessionId + " " + usrid; 
	 else
	     reply = parseUserDatabase(reply, "login");	
         return reply;
    }

    public String changeRegInfo(String msg) {
        /* This method is called to change existing registration info.
	   Msg format is:
		oldnick [Real Name] [email@oldaddress] [old company] \
	        [old title]|newnick newpassword [Real Name] [email@newaddr] \
		[new company] [new title] [collab groups] [optional info]
	*/
	if (DEBUG)
	    System.out.println("ClientThread(" + usrid + 
		")::" + new Date() + ": changeRegInfo()--msg is: " + msg);
	int i = msg.indexOf("["); 
	if (i < 0)
	    return "changereg BADREQUEST";
 	String oldNick = msg.substring(0, (i-1));
	msg = msg.substring(i); 
	i = oldNick.indexOf("_");
	if (i > -1)
	    oldNick = oldNick.substring(0, i);
	String usridStripped = usrid;
	i = usridStripped.lastIndexOf("_");
	if (i > -1)
	    usridStripped = usridStripped.substring(0, i);
	if (usrid.startsWith("guest") || 
	   (!usridStripped.equals(oldNick) && nickChange==null) || 
	   (nickChange!=null && !oldNick.equals(nickChange)))
	       return "changereg NOTAUTHORIZED";	
	StringTokenizer t = new StringTokenizer(msg, "|"); 
	if (t.countTokens() < 2)
	    return "changereg BADREQUEST";
	String oldInfoSent = t.nextToken();
	String newInfoSent = t.nextToken();
	
	i = newInfoSent.indexOf(" ");
	if (i < 0)
	    return "changereg BADREQUEST";
	String newnick = newInfoSent.substring(0, i);
	if (!newnick.equals(oldNick))
	{
	    if (fileHandler.findUser(newnick) != null)
	        return "changereg NICKINUSE " + newnick;
	}
	i++;
	newInfoSent = newInfoSent.substring(i); 
	i = newInfoSent.indexOf(" ");
	if (i < 0)
	    return "changereg BADREQUEST";
	String newPassword = newInfoSent.substring(0, i);
	newInfoSent = newInfoSent.substring(++i);
        String dbEntry = fileHandler.findUser(oldNick);
	t = new StringTokenizer(dbEntry);
	t.nextToken();
	String oldPassword = t.nextToken();

	// If user sent "*", use the old password. 
	if (newPassword.equals("*"))
	    newPassword = oldPassword;	
	// change 'user' db entry
	try {
	    fileHandler.replace("user " + dbEntry, 
		"user " + newnick + " " + newPassword + " " + newInfoSent); 
	} catch(Exception e) {
	    return "changereg FAILED";
	}
	// change nicks for all registered certs
	if (!newnick.equals(oldNick))
	{
	    try {
		String[] certs = fileHandler.findAllByKeys("dn", oldNick);
		for (i=0; i<certs.length; i++)
		{
		    String newString = "dn " + newnick + " " +
			certs[i].substring(oldNick.length()+1);
		    fileHandler.replace("dn " + certs[i], newString);
		}
	    } catch(Exception e) {
		return "changereg FAILED";	
	    }
	} 
	nickChange = newnick;

	StringTokenizer t2 = new StringTokenizer(newInfoSent, "[");
	String temp = null;
	if (t2.hasMoreTokens())
	{
	    temp = t2.nextToken();
	    temp.trim();
	    if (temp.endsWith("]"))
		temp = temp.substring(0, temp.length()-1);
	    realname = temp.trim();
	    thisUser.setRealname(realname);
	}
	if (t2.hasMoreTokens())
	{
	    temp = t2.nextToken();
	    temp.trim();
	    if (temp.endsWith("]"))
		temp = temp.substring(0, temp.length()-1);
	    emailAddr = temp.trim();
	    thisUser.setEmail(emailAddr);
	}
	if (t2.hasMoreTokens())
	{
	    temp = t2.nextToken();
	    temp.trim();
	    if (temp.endsWith("]"))
		temp = temp.substring(0, temp.length()-1);
	    company = temp.trim();
	    thisUser.setCompany(company);
	}
	if (t2.hasMoreTokens())
	{
	    temp = t2.nextToken();
	    temp.trim();
	    if (temp.endsWith("]"))
		temp = temp.substring(0, temp.length()-1);
	    title = temp.trim();
	    thisUser.setTitle(title);
	}
	if (t2.hasMoreTokens())
	{
	    temp = t2.nextToken();
	    temp.trim();
	    if (temp.endsWith("]"))
		temp = temp.substring(0, temp.length()-1);
	    groups = temp.trim();
	    if (!groups.equals(thisUser.getGroupsAsString()))
	    {
		thisUser.clearGroups();
		StringTokenizer t3 = new StringTokenizer(groups);
		while (t3.hasMoreTokens())
		    thisUser.addGroup(t3.nextToken());
	    }
	}
	if (t.hasMoreTokens())
	{
	    temp = t.nextToken();
	    temp.trim();
	    if (temp.endsWith("]"))
		temp = temp.substring(0, temp.length()-1);
	    thisUser.setDescription(temp);
	}
	
	dbEntry = dbEntry.substring(
	    dbEntry.indexOf(oldPassword) + oldPassword.length() + 1); 
	if (!oldNick.equals(newnick) || !newInfoSent.equals(dbEntry))
	    server.sendAll("changereginfo SUCCESS " + oldNick + " " + newnick 
		+ " " + newInfoSent);
	else if (!newPassword.equals(oldPassword))
	    return("changepassword SUCCESS " + newPassword);
	return("changereg SUCCESS " + newnick + " " + newPassword + " " +
	    newInfoSent);
    }

    private String newRegistration(String msg)
    {
	/* Called to register another user.  The sender must have established an
	   SSL connection and presented an X.509 valid certificate.
	   Msg in format:
		sendernick usernick userpassword [User's Full Name] \
		[User's Email Address] [User's Affiliation] [User's Job] \
	   	[Collaboration Groups] [Optional description]  
	*/  
	if (DEBUG)
	    System.out.println("ClientThread(" + usrid + 
		")::" + new Date() + ": newRegistration()--msg: " + msg); 
	StringTokenizer t = new StringTokenizer(msg);
	if (t.countTokens() < 4)
	    return "newreg BADREQUEST";
	if (!authenticated)
	    return "newreg NOTAUTHORIZED"; 	
	// strip off sender's name
	t.nextToken();
	String usernick = t.nextToken();
	String userpass = t.nextToken(); 
	t = new StringTokenizer(msg, "[");
	if (t.countTokens() < 4) 
	    return "newreg BADREQUEST";
	String userinfo = msg.substring(msg.indexOf("["));			
	
	/* Check if user is already registered or the nick is being used by 
           someone else.
	*/
  	String reply = fileHandler.findUser(usernick);	
	if (reply != null)
	{
	    reply = reply.substring(reply.indexOf("["));	
	    if (userinfo.equals(reply))
	        return "newreg ISREGISTERED " + usernick; 
	    else
		return "newreg NICKINUSE " + usernick;
	}
	try {
	    fileHandler.writeToFile("user " + usernick + " " + userpass + " " + 
		userinfo);		
	} catch(Exception e) {
	    return "newreg FAILED " + usernick;
	}
	server.sendAll("newreg " + usernick + " offline false " + userinfo);
	return("newreg SUCCESS " + usernick); 
    }	

    public String doRegister(String msg) 
    {
	/* This method is called when a new user wants to register a
	   permanent userid (which will be a permanent 'nick' for IRC).
	   Assumptions: (1) one real user can have multiple 'nicks';
	   (2) each 'nick' must be unique (e.g., Marcia Perry at LBL 
	   with email address mperry@lbl.gov can register with userids 
	   'marcia' and 'mperry', but no one else can have either of 
	   these nicks if they are previously assigned.); (3) email
	   address, affiliation, and title are optional fields, but
	   a 'real name' must be supplied; (4) msg format is:
		tool sessionId senderHost usrid passwd [Real Name] \
		[email@addr] [company name] [job title] 

           Algorithm: Check that there are at least 4 fields (tool, usrid,
	   passwd, realname). Assuming ok, proceed as follows:  
	   Search the database for a 'user' entry for 'usrid'. If
	   none is found, write the msg to the database and send
	   SUCCESS.  If there is an entry for 'usrid' in the database,
	   compare the password in the database w/ that in the message.
	   If they are the same, assume this is a re-registration so just
	   send ISREGISTERED reply.  If the passwords don't match, the
	   user must pick a different userid, so send NICKINUSE.
        */
	     
	if (DEBUG)
	    System.out.println("ClientThread(" + usrid + 
		")::" + new Date() + ": doRegister()--msg is: " + msg);
	msg = parseHeader(msg);
	if (msg == null || msg.equals("BADHEADER")) 
	     return("register BADREQUEST");

	String reply = fileHandler.findUser(usrid);
	if (reply == null)
	{
	     /* This user was not found (and therefore couldn't be logged in from
	        another session), so write the entry to the database and and treat 
   		as a login.
 	     */
	     int i = msg.indexOf("[");
	     if (i < 0)
		return "register BADREQUEST";
	     try {
	        fileHandler.writeToFile("user " + usrid + " " + passwd + " " + msg);	
	     } catch(IOException e) {
		return "register FAILED";
	     }
	     reply = "register SUCCESS " + sessionId + " " + usrid; 
	     parseUserInfo(msg);
	     server.addSessionId(this, sessionId);
	     server.addUser(sessionId, createUser(), tool);
	}
	else
	     reply = parseUserDatabase(reply, "register");
	return reply;
    }

    // This method sets the availability status for the user.
    private String doStatus(String msg)
    {
	String reply = "NONE";
	msg = parseHeader(msg);
	if (msg != null)
	{
	    if (msg.equals("BADHEADER"))
		reply = "status BADREQUEST";
	    else
		server.setUserStatus(sessionId, tool, msg, this);
	}
	else
	    reply = "status BADREQUEST";
	return reply;
    }

    // This method sets the user's online location (e.g., work home, mobile).
    private String doLocation(String msg)
    {
	String reply = "NONE";
	msg = parseHeader(msg);
	if (msg != null)
	{
	    if (msg.equals("BADHEADER"))
		reply = "location BADREQUEST";
	    else
		server.setUserLocation(sessionId, tool, msg, this);
	}
	else
	    reply = "location BADREQUEST";
	return reply;
    }

    // This method sets the channel to 'open' or 'closed'.
    private String doDoor(String msg)
    {
	String reply = null;
	msg = parseHeader(msg);
	if (msg != null)
	{
	    if (msg.equals("BADHEADER"))
		reply = "door BADREQUEST";
	    else
	    {
		StringTokenizer t = new StringTokenizer(msg);
		if (t.countTokens() < 2)
		    reply = "door BADREQUEST";
		String name = t.nextToken(); 
		boolean closed = true;
		if (t.nextToken().equals("O"))
		    closed = false;
		server.setDoor(sessionId, name, tool, closed, this);
	    }
	}
	return reply;
    }

    private String parseHeader(String s)
    {
	/* login and register requests have the following header:
		tool sessionId senderHost nick passwd
	   Extract the fields, set private variables, and return the remaining portion
	   of the String 's'.
	*/    
	StringTokenizer t = new StringTokenizer(s);
	if (t.countTokens() < 5)
	    return "BADHEADER";
	tool = t.nextToken();
	sessionId = t.nextToken();
	senderHost = t.nextToken();
	usrid = t.nextToken();
	passwd = t.nextToken();
	int len = tool.length() + sessionId.length() + senderHost.length() + 
	    usrid.length() + passwd.length() + 5;
	if (t.hasMoreTokens())
	    return s.substring(len);
	return null;
    }

    private int parseUserInfo(String s)
    {
	/* guest_login and register requests have this content after the header:
		[Real Name] [email@address] [company or affiliation:division] \
		[job title] [collaboration groups]

	   These are also the last fields of database 'user' entries. 
	   Extract these fields and set private variables.
        */ 
	StringTokenizer t = new StringTokenizer(s, "[");
	int numTokens = t.countTokens();
	if (numTokens < 4)
	    return 0;
	realname = t.nextToken().trim();
	if (realname.endsWith("]"))
	    realname = realname.substring(0, realname.length()-1);
	emailAddr = t.nextToken().trim();
	if (emailAddr.endsWith("]"))
	    emailAddr = emailAddr.substring(0, emailAddr.length()-1);
	company = t.nextToken().trim();
	if (company.endsWith("]")) 
	    company = company.substring(0, company.length()-1);
	title = t.nextToken().trim();	
	if (title.endsWith("]"))
	    title = title.substring(0, title.length()-1);
	if (t.hasMoreTokens())
	{
	    groups = t.nextToken().trim();
	    if (groups.endsWith("]"))
	        groups = groups.substring(0, groups.length()-1);
	}
	return numTokens;
    }
   
    private String parseUserDatabase(String s, String type)
    {
        /* 'type' denotes 'login' or 'register'; 'pass' is the password.
	   's' is in the form: 
	     userid passwd [realname] [email] [company:optional division] \
	     [title] [collaboration groups] 
	   e.g.,
	     marcia 1234 [Marcia Perry] [mperry@lbl.gov] [Some company name] \
	     [DSD Staff] [PCCE DIDC]
	*/

	String reply = type;
	int i = s.indexOf("[");
	// If no '[' is found, it's a bad database entry.
	if (i <= 0)
	    return(type + " NOTREGISTERED");
		
	int l = s.length(); 
	
	StringTokenizer t = new StringTokenizer(s);
	// Strip off the userid.
	t.nextToken();	
	String pass = t.nextToken();	
	if (passwd.equals(pass))
	{
	   s = s.substring(i,l);
	   if (parseUserInfo(s) == 0)
	   {
	      if (type.equals("login"))
		return("login NOTREGISTERED " + usrid + " " + usrid);
	      else if (type.equals("register"))
		return("register BADREQUEST");	
	   }
	   if (type.equals("login"))
	   {
	      usrid = server.getUserid(usrid);	
	      sessionId = usrid + senderHost;
	      reply = reply + " SUCCESS " + sessionId + " " + usrid;
	   }
           else if (type.equals("register"))
	      reply = reply + " ISREGISTERED " + usrid; 

	   server.addSessionId(this, sessionId);

	   /* If this is a new login or registration, add the user and set all the 
	      information; otherwise, change the nickname and password in the existing 
	      User object.
	   */  
	   server.addUser(sessionId, createUser(), tool);
	   reply = reply + " " + getState(); 
	}	
	else
	{
	   if (type.equals("login"))
	      return("login BADPASSWD " + usrid);
	   else if (type.equals("register"))
	      return("register NICKINUSE " + usrid);
	}
	return reply;
    }

    /* Called to confirm that request received over this connection was sent by
       the same user who was authenticated by an X.509 certificate when the connection 
       was established.
    */ 
    private boolean checkSender(String sendernick)
    {
	User sender = server.getUser(this);
	String realname = sender.getRealname();
	Vector peerCerts = myComm.getPeerDN();
	if (peerCerts == null)
	    return false;
	StringTokenizer t = null;
	for (int i=0; i<peerCerts.size(); i++)
	{
	    String DN = (String)peerCerts.elementAt(i);
	    t = new StringTokenizer(DN, ",");
	    while (t.hasMoreTokens())
	    {
		String field = t.nextToken().trim();
		if (field.toLowerCase().startsWith("cn"))
		{
		    int indx = field.indexOf("=");
		    if (indx > 0)	
		    {
			field = field.substring(indx+1);
			if (field.equals(realname))
			    return true;
		    }	
		}
	    }
	} 
	return false;
    }

    public String getTool() { return tool; }
	
    public boolean sendMsg(String msg) 
    {
	if (DEBUG)
	    System.out.println("ClientThread(" + usrid + ")::" + new Date() + 
	 	": sendMsg calling writeMessage()--msg: " + msg);
	boolean sent = false; 
	try {
	    sent = myComm.writeMessage(msg);
	    System.out.println("ClientThread(" + usrid +
	       ")::" + new Date() + ": sendMsg() writeMessage() ret'd: " +sent);
	} catch(SocketException e) {
	    System.out.println("ClientThread(" + usrid + ")::" + new Date() + 
		": sendMsg() writeMessage() SocketException: " +e.getMessage());
	    return sent;
	} catch(IOException e) {
	    System.out.println("ClientThread(" + usrid + ")::" + new Date()
	   	+ ": sendMsg() writeMessage() IOException: "+e.getMessage());
	    return sent;
        } catch(Exception e) {
            System.out.println("ClientThread(" + usrid + ")::" + new Date()
	   	+ ": sendMsg() writeMessage() Exception: " + e.getMessage());
	    return sent;
        }
        return sent;
    }	

    private String doGet(String keyword)
    {
	if (keyword.equals("pchan") || keyword.equals("tchan"))
	    return server.getChannels(keyword);
	else if (keyword.equals("mcastaddrs"))
	    return server.getMulticastAddrs();
	else if (keyword.equals("allusers"))
	    return getEntries("user", "allusers");
	else if (keyword.equals("note"))
	    return(server.getNotes(thisUser));
	else if (keyword.equals("state"))
	    return getState();
	else if (keyword.toLowerCase().startsWith("pcce") ||
	    keyword.toLowerCase().startsWith("vers"))
		return "pcceversion " + PCCEServer.version;
	else
	    return getEntries(keyword);
    }

    private String doPut(String args)
    {
	int i = args.indexOf(" ");
	if (i < 0)
	    return null;
	String cmd = args.substring(0, i).toLowerCase();
	args = args.substring(i+1);
	if (cmd.startsWith("note"))
	    return server.putNote(usrid, args);
	else if (cmd.equals("cert"))
	{
	    i = args.indexOf(" ");
	    if (i < 0)
		return "putcert BADREQUEST";
	    String nick = args.substring(0, i);
	    args = args.substring(i+1);
	    i = nick.lastIndexOf("_");
	    if (i > -1)
		nick = nick.substring(0,i);
	    String value = nick + " " + args; 
	    String[] entry = fileHandler.findAllByKeys("dn", value);
	    if (entry != null)  
	    {
		for (i=0; i<entry.length; i++)
		{
		    if (entry[i].equals(value))
			return "putcert SUCCESS";
		}
	    } 
	    try {
	        fileHandler.writeToFile("dn " + value);
	    } catch(Exception e) { }
	    return "putcert SUCCESS";
	}
	return null;
    }

    private String doDelete(String msg)
    {
	if (msg.startsWith("note"))
	    thisUser.deleteNote(msg);
	return null;    
    }

    private String getState()
    {
	StringBuffer msg = new StringBuffer("state SUCCESS voname ");
	String[] entry = fileHandler.findAllByKey("voname");	
	if (entry != null)
	    msg.append(entry[0]); 
	String s = server.getChannels();
	if (s != null && s.length() > 0)
	    msg.append("|" + s);
	entry = fileHandler.findAllByKey("securechannel");
	if (entry != null)
	    msg.append("|securechannel " + entry[0]);
	s = getUsers("registered");
	if (s != null && s.length() > 0)
	    msg.append("|" + s);
	s = server.getToolUsers(tool);
	msg.append("|" + s);	
	s = thisUser.getNotes();
	if (s != null && s.length() > 0)
	    msg.append("|" + s);
	return msg.toString();
    }
 
    private String getUsers(String type)
    {
	StringBuffer buff = new StringBuffer();
	if (type.toLowerCase().startsWith("reg"))
	{
	    String[] entries = fileHandler.findAllByKey("user");
	    if (entries != null)
	    {
		for (int i=0; i<entries.length; i++)
		{
		    StringTokenizer t = new StringTokenizer(entries[i]);	
		    buff.append("reguser " + t.nextToken() + " offline false ["); 
		    int indx = entries[i].indexOf("[");
		    if (indx > -1) 	
			buff.append(entries[i].substring(indx+1) + "|");
		    else
			buff.append("[null]|");
		}
	    }	
	}
	String reply = buff.toString().trim();
	if (reply.endsWith("|"))
	    reply = reply.substring(0, reply.length()-1); 
	return reply;
    }

    private String getEntries(String keyword, String msgType)
    {
	/* This method returns all the entries in the database for the
           given keyword.
	*/
	String[] entries = fileHandler.findAllByKey(keyword);
	int indx; 
	if (entries != null)
	{
	    StringBuffer buff = new StringBuffer(msgType + " SUCCESS ");
	    for (int i=0; i<entries.length; i++)
	    {
		buff.append(entries[i]); 
		buff.append("|");
	    }	
	    String reply = buff.toString();
	    return reply.substring(0, reply.length()-1);	
	} 
	return null;
    }

    private String getEntries(String key)
    {
	String[] entries = fileHandler.findAllByKey(key);
	int indx;
	if (entries != null)
	{
	    StringBuffer buff = new StringBuffer(key + " SUCCESS ");
	    for (int i=0; i<entries.length; i++)	
	    {
		buff.append(entries[i]); 
		buff.append("|");
	    }
	    String reply = buff.toString();
	    return reply.substring(0, reply.length()-1);
	} 
	return null;	
    }

    private String forwardJoinReq(String msg)
    {
	StringTokenizer t = new StringTokenizer(msg);
	if (t.countTokens() < 2)	
	    return "BADREQUEST";	
	String id = t.nextToken();
	if (server.sendMsg("join " + t.nextToken(), id) == true)
	    return "join SUCCESS";
	else
	    return null;	
    }

    private String forwardPingReq(String id)
    {
	if (server.sendMsg("ping", id) == true)
	    return "PONG";
	return null;
    }		

    // Methods for IRC support.
    private String newVenue(String msg)
    {
	msg = parseHeader(msg);	
	StringTokenizer t = new StringTokenizer(msg);
	if (t.countTokens() < 4)
	    return("newvenue BADREQUEST");
	String name = t.nextToken();
	String reply = "newvenue SUCCESS " + name;
	String creator = t.nextToken();
	boolean perm = true;
	boolean closed = true;
	if (t.nextToken().equals("T"))
	    perm = false;
	if (t.nextToken().equals("O"))
	    closed = false;
	String desc = null;
	if (t.hasMoreTokens())
	{
	    int startIndx = msg.indexOf("(");
	    int endIndx = msg.indexOf(")");
	    if (startIndx > -1)
	    {
		if (endIndx > -1)
		    desc = msg.substring(startIndx, endIndx);
		else
		    desc = msg.substring(startIndx);
	    }
	    else
		desc = t.nextToken();
	}
	server.addChannel(sessionId, tool, this, name, desc, creator, perm, closed); 
	return reply;
    }

    public void closeConnection() { myComm.cleanup(); }
		
    protected void finalize() 
    {
	myComm.cleanup();
	myComm = null;
    }
}
