import pccejava.*;
import java.net.*;
import java.io.*;
import java.util.*;

public class IRCClient implements Runnable {
    protected String host = "shaggy.lbl.gov";
    protected int port = 7000;
    protected SSLComm comm = null;
    protected String localaddr = null;
    protected PCCEServer server = null;
    protected String keyfile = "pcce.p12";
    protected String casfile = "certs.pem";
    protected String passphrase = "remake28";	
    protected InetAddress serverAddr = null; 
    private static boolean DEBUG = true;
    private int numPongs = 0;
    public static boolean IRCconnect = true;

    public IRCClient(PCCEServer caller, String h, int p)
    {
	server = caller;
	host = h;
	port = p;
    }

    public void setSecurityInfo(String key, String cas, String pass)
    {
	keyfile = key;
	casfile = cas;
	passphrase = pass;
    }

    public void run() 
    {
	String reply = "";
	comm = new SSLComm(host, port, keyfile, casfile, passphrase);
	try {
	    serverAddr = InetAddress.getByName(host);
	} catch(UnknownHostException e) {
	    System.err.println("PCCEServer can't get hostname of IRC server");
	    IRCconnect = false;
	    comm = null;
	    return;	
	}
	try {
	    comm.initSSL();
	    comm.connectSocket();
	    comm.setSocketTimeout(540000);
	} catch(Exception e) {
	    System.err.println("PCCEServer can't connect to IRC Server");
	    e.printStackTrace();
	    IRCconnect = false;
	    comm.cleanup();
	    comm = null;
	    return;	
	}	
	localaddr = comm.getSSLAddress().toString();
	localaddr = localaddr.substring(1);
	try {
	    comm.setIOStreams();
	} catch(Exception e) {
	    System.out.println("PCCEServer can't set IOStreams for IRCD");
	    e.printStackTrace();
	    IRCconnect = false;
	    comm.cleanup();
	    comm = null; 
	    return;	
	} 
	comm.sendMessage("NICK pcceserve " + PCCEServer.ircPasswd);
	comm.sendMessage("USER pcceserve " + localaddr + " " +
	    serverAddr + ": PCCEServer");
	comm.sendMessage("JOIN #Directory");
	IRCconnect = true;
	
	while (true)
	{		
	    boolean sent; 
	    try {
	 	reply = comm.getSSLMessage();
		if (DEBUG)
	           System.out.println("IRCClient(" + new Date() + ") RCV'D: " + 
			reply);
	    } catch(SocketTimeoutException e) {
		System.out.println("IRCClient(" + new Date() + 
		    ")::socket timed out");
		comm.cleanup();
		comm = null;
		return;
	    } catch(IOException e) {
		System.out.println("IRCClient(" + new Date() + 
		    ")::IOException: " + e.getMessage());
	    } catch(Exception e) { 
		System.out.println("IRCClient(" + new Date() + 
		    ")::can't get messages from IRCD");
		e.printStackTrace();
		comm.cleanup();
		comm = null;
		return;
	    }
	    String prefix = IRCParser.getPrefix(reply);
	    String sender = IRCParser.getSender(prefix);
	    String cmd = IRCParser.getCmd(reply).toLowerCase();
	    if (prefix.toLowerCase().equals("ping"))
	    {
		String msg = "PONG " + IRCParser.getArgs(reply, 2); 
		sent = comm.sendMessage(msg);
		if (DEBUG)
		    System.out.println("IRCClient(" + 
		        new Date() + ")::wrote PONG--sent: " + sent + " msg: " + msg);
		numPongs++;
		if (numPongs == 10)
		{	
		    comm.sendMessage("NAMES"); 
		    numPongs = 0;	
		}
	    }
	    else if (cmd.equals("quit"))
		server.killConnection(sender);
	}
    }
}	
