import java.awt.*;

// This class parses messages from the IRC server.

public class IRCParser {

    /**
     * This method is used to extract the prefix from which we can identify the
     * sender.
     * The leading semicolon (:) is not returned.
     * @param message the message to be parsed
     * @return the leading part of the message up to the first space character or 
     * the message itself (whichever comes first).
     */
    public static String getPrefix(String message)
    {
	if (message == null)
	    return null;
	else
	{
	    if (message.indexOf(" ") > -1) 
		message = message.substring(0, message.indexOf(" "));
	}
	if (message.charAt(0) == ':')
	    return message.substring(1);
	else
	    return message; 
    }
   
    public static String getArgs(String message, int n)
    {
	for (int i=0; i<n-1; i++)
    	{
	    if (message.indexOf(" ") == -1) 
		return null;
	    else 
		message = message.substring(message.indexOf(" ") + 1);
	}
	return message;
    }

    public static String getCmd(String message)
    {
	if (message == null) return null;
	else
	{
	    if (message.indexOf(" ") == -1) 
		return message;
	    else 
		message = message.substring(message.indexOf(" ") + 1);
	    if (message.indexOf(" ") == -1) 
		return message;
	    else 
		return message.substring(0, message.indexOf(" "));
	}
    }

    /**
     * Use only for N > 2.
     * @param message - the message to be parsed
     * @param N       - the index specifying which part of the message, 
     *                  separated by space characters, is to be extracted.
     * @return the Nth part or the last part of the message (whichever comes first).
     */
    public static String getNth(String message, int n)
    {
	if (message == null) 
	    return null;
	else
	{
	    for (int i=0; i<n-1; i++)
	    {    
	        if (message.indexOf(" ") == -1) 
		    return message;
	        else 
		    message = message.substring(message.indexOf(" ") + 1);
	    }    
	    if (message.indexOf(" ") == -1) 
	    	return message;
	    else 
	    	return message.substring(0, message.indexOf(" "));
	}
    }
 
    /* This method extracts the sender from the prefix, in the form: 
       <nick>[!<nick>[@<host], e.g., marcia!marcia@myhost.lbl.gov		
		        or
       <sender_host>		
    */
    public static String getSender(String prefix)
    {
	if (prefix == null) 
	    return null;

	int i = prefix.indexOf("!");
	if (i > -1) 
	    return prefix.substring(0, i);
	else 
	    return prefix; 
    }
}
