/* This is the server for the Pervasive Collaborative Computing Environment.  It should be run on the web server so that  applets can connect to it.
*/

import pccejava.*;
import java.net.*;
import java.util.*;
import java.io.*;
import javax.net.ssl.*;

class PCCEServer {
    private boolean DEBUG = true;

    // network communication object for use by this (main) thread
    private SSLComm comm = null;  

    // database for permanent collaboration objects
    private static String database = "pcce.db"; 
    private FileHandler fh = null;	

    /* hashtable of userid objects; key is a login userid (a 'nickname' w/o a 
       numerical suffix); value is a Userid object (which represents a nickname 
       prefix)
    */
    private Hashtable nicks;

    // hashtable of user objects; key is a 'sessionId'; value is a User object  
    private Hashtable users;

    // hashtable of open connections: key is 'sessionId'; value is a ClientThread object
    private Hashtable conns;

    // hashtable of chat channels; key is 'channel name'; value is a Channel object
    private Hashtable channels;
	
    private static String keyFile = "pcce.p12";
    private static String casFile = "certs.pem";
    private static String password = "remake28";
    public static String version = "2.2";
  
    // For verifying msg is from IRCD 
    public static String ircPasswd = "ircd!pass";

    // For exchanging msgs w/ IRC server.
    private static String ircHost = "pcce.lbl.gov";
    private static int ircPort = 7000;

    public static void main(String[] args) 
    {
	for (int i=0; i<args.length; i++)
	{
	    if (args[i].equals("-f"))
		database = args[++i];
	    else if (args[i].equals("-p"))
	        password = args[++i];
	    else if (args[i].equals("-k"))
		keyFile = args[++i];
	    else if (args[i].equals("-c"))
		casFile = args[++i];  
	    else if (args[i].equals("-ircpass"))
	        ircPasswd = args[++i];
	    else if (args[i].equals("-irchost"))
		ircHost = args[++i];
	    else if (args[i].equals("-ircport"))
		ircPort = Integer.parseInt(args[++i]);	
	    else if (args[i].startsWith("-v"))
	    {
		i++;
		System.out.println("PCCEServer version " + version);	
	    }
	} 
	new PCCEServer().start();
    }

    public void start() 
    {
        nicks = new Hashtable();
	users = new Hashtable();
	conns = new Hashtable();
        channels = new Hashtable();

	/* Instantiate a NetComm object for network communication. Get a 
           'main' socket on which to listen for client connection requests.  
           When a connection is accepted, create a ClientThread to handle that 
           client. TCP connections are held open until the client closes it. 
        */

	if (keyFile!=null && casFile!=null && password!=null)
	{
            comm = new SSLComm(null, 7310, keyFile, casFile, password);
	    System.out.println("PCCEServer getting Server Socket");
	    try {
		comm.initSSL();
                comm.openServerSocket();
	    } catch(Exception e) { 
	        System.out.println("Can't listen for connection requests...terminating" + 
		    e.getMessage()); 
	        System.exit(1);
	    }
	    System.out.println("PCCEServer READY");

	    /* Get a FileHandler so the client threads can access the database
	       in a synchronized way.  
            */
	    fh = new FileHandler(database);

            /* Cache the permanent channels so we don't have to do file I/O every time a 
               user asks for them.
    	    */
	    initializeChannels();

	    // Connect to IRC server.  Maybe the archiver can do this? 
	    if (DEBUG)
	        System.out.println("PCCEServer::instantiating IRCClient--ircHost: "
		    + ircHost);
	    IRCClient irc = new IRCClient(this, ircHost, ircPort);
	    irc.setSecurityInfo(keyFile, casFile, password);
	    ThreadMgr thrMgr = new ThreadMgr(irc);
	    Thread mgrThread = new Thread(thrMgr, "ThreadMgr");
	    mgrThread.start();
	    /*
	    Thread ircThr = new Thread(irc, "IRCClient");
	    ircThr.start(); 	
	    */
 
            //Listener listener = new Listener(this, fh);
	    //listener.start();

	    for (;;) {
	        ClientThread conn = connectToClient(); 
	        if (conn != null) {
		    if (DEBUG)
		       System.out.println("PCCEServer::connectToClient() returned conn "
		           + conn);
	        }
  	    }
	}
    }

    public String getDatabase() { return database; } 

    private void initializeChannels()
    {
	String[] entries = fh.findAllByKey("channel");
	StringTokenizer t = null;
	String name = null, temp = null;
	if (entries != null)
	{
	    for (int i=0; i<entries.length; i++)
	    {
		t = new StringTokenizer(entries[i]);
		if (t.countTokens() < 3)
		    return;
		name = t.nextToken();
		temp = t.nextToken();
		String desc = null;
		int indx = entries[i].indexOf("("); 
		if (indx > -1)
		    desc = entries[i].substring(indx+1);
		else
		    desc = t.nextToken();		
		indx = desc.indexOf(")");
		if (indx > -1)
		    desc = desc.substring(0, indx);
		channels.put(name, new Channel(name, desc, temp, true, false));
	    }
	}
	String[] addrs = fh.findAllByKey("mcast");
	if (addrs != null)
	{
	    for (int i=0; i<addrs.length; i++)
	    {
		t = new StringTokenizer(addrs[i]);
		if (t.countTokens() < 3)
		    return;
		name = t.nextToken();
		temp = t.nextToken(); 
		int port = Integer.parseInt(t.nextToken());
		Channel ch = (Channel)channels.get(name);
	        if (ch !=null)
	 	{
		    ch.setMcastAddress(temp);
		    ch.setMcastPort(port);
		}
	    }
	}
    }

    // This is called when ircd wants to check access; will be replaced by Akenti.
    public String doIRCServer(String msg)
    {
	String nick=null, passwd=null, realname=null, reply=null;
	StringTokenizer t = new StringTokenizer(msg);
	if (t.countTokens() < 2)
	    return("BADREQUEST"); 
	if (!t.nextToken().equals(ircPasswd))
	    return("NOTAUTHORIZED"); 
	if (t.nextToken().toLowerCase().equals("online"))
	{
	    nick = t.nextToken();
	    passwd = t.nextToken();
	    if (nick==null || passwd==null)
		return("online BADREQUEST");
	    int i = msg.indexOf("[");
	    if (i > -1)
	    {
		realname = msg.substring(i+1);
		if (realname.endsWith("]"))
		    realname = realname.substring(0, realname.length()-1);
	    }	
	    String key = null;
	    User usr = null;
	    synchronized(users) {
	        if (users.isEmpty())
		    reply = "online FAILURE " + nick;
		else
	        {
	            for (Enumeration e=users.keys(); e.hasMoreElements();)
	            {
			key = (String)e.nextElement();
			usr = (User)users.get(key);
			if (usr!=null && usr.isRunning("chat") && 
		    	    usr.getNickname().equals(nick)) 
			{
		    	    if (usr.authenticated() && realname!=null && 
				usr.getRealname().equals(realname))
		              reply = "online SUCCESS " + nick;
		            else if (!usr.authenticated() && usr.getPasswd().equals(passwd))
			        reply = "online SUCCESS " + nick; 
			    break;
		        }	
	            }
	        }
	    }
	}
	return reply;
    }

    public synchronized String getNick(String id) 
    {
    	/* This method returns the nickname in the User object that is associated
	   with the given 'id', which is a sessionId.
        */
	User usr = (User)users.get(id);
	if (usr == null)	
	    return null;
	return usr.getNickname();
    }

    public String getUserid(String name) 
    {
	/* This method will be called by the threads (when a 'login' or 'register' request 
           is received) to get a unique nickname for the IRC server.  If the 'nicks' 
           hashtable has this name, get a new nickname with an incremented suffix 
           appended to it; otherwise, create a Userid object and add it to the hashtable.   
	*/ 

	Userid uid;
	synchronized(nicks) {
	    if (nicks.containsKey(name))
	        uid = (Userid)nicks.get(name);
	    else
	    {
	        uid = new Userid(name);
	        nicks.put(name, uid);
	    }
	}
	return(uid.generateNick());
    }

    public void deleteUserid(String nick) 
    {
   	/* Called when a user logs out.  If 'nick' contains a numerical
	   suffix (e.g., 'guest_2'), strip off the suffix.  Delete the
	   result from the hashtable.
	*/
	int index = nick.indexOf("_");
	Integer i = new Integer(1);
	String name = nick;
	if (index > -1)
        {
	    name = nick.substring(0, index);
	    String suffix = nick.substring((index+1),nick.length());
	    i = Integer.valueOf(suffix);
	} 
	synchronized(nicks) {
 	    if (nicks.containsKey(name))
	    {
	        Userid uid = (Userid)nicks.get(name);
	        uid.deleteNick(i);
	        if (uid.isEmpty())
		    nicks.remove(name); 
	    }	
	}
    }

    public synchronized boolean sessionExists(String id) 
    {
	return users.containsKey(id);
    }

    // Return a list of hosts on which the user is running a given tool
    public String[] toolRunning(String usrid, String tool,
	String realname)
    {
	Vector v = new Vector();
	String key;
	User usr;
	
	synchronized(users) {
	    for (Enumeration e=users.keys(); e.hasMoreElements(); )
            {
	        key = (String)e.nextElement();
	        usr = (User)users.get(key);
	        if (usr != null)
	        {
		    if (usr.getNickname().startsWith(usrid) &&
		        usr.getRealname().equals(realname) && usr.isRunning(tool))
			    v.add(usr.getHost());
	        }
	    }
	}
	int len = v.size();
	String hosts[] = new String[len];
	for (int i=0; i<len; i++)
	    hosts[i] = (String)v.elementAt(i);
	return hosts;
    }

    // Add a User object to the hashtable and notify everyone online. 
    public synchronized void addUser(String id, User usr, String app)
    {
	users.put(id, usr);
	for (Enumeration enum=conns.keys(); enum.hasMoreElements();)
	{
	    ClientThread thr = (ClientThread)enum.nextElement();
	    thr.sendMsg("newuser SUCCESS " + app + " " + usr.getUserInfo() 
		+ " [" + usr.getGroupsAsString() + "]");
	} 
    }

    /* This method is called when a user registers, logs in, or starts a new 
       tool to add the user to the Users hashtable (or add a tool to an
       existing user).
    */
    public synchronized void addUser(String id, String n, String p, String r, 
	String e, String c, String t, String h, String app, boolean auth) 
    {
	/* If a User object exists, add this app as a tool; else
	   instantiate a User object and set the user information.
        */ 
	User usr;
	if (users.contains(id))
	{
	    usr = (User)users.get(id);
	    usr.addTool(app);
	}
	else
	{
	    usr = new User(id);
	    usr.setUserInfo(n, p, r, e, c, t, h, app);	

	    // Add this User object to the hashtable.
	    users.put(id, usr);
	}
	usr.setAuthenticated(auth);
	    
	// Notify all the other users.
	for (Enumeration enum=conns.keys(); enum.hasMoreElements(); )
	{
	    ClientThread thr = (ClientThread)enum.nextElement();
	    if (!conns.get(thr).equals(id))
		thr.sendMsg("newuser SUCCESS " + app + " " + usr.getUserInfo());
	}
    } 

    public synchronized void deleteUser(String id) { users.remove(id); }

    public void setPassword(String pw) { password = pw; }

    public void setp12file(String p12) { keyFile = p12; }

    public void setpemfile(String pem) { casFile = pem; }

    public synchronized void doLogout(String id)   
    {
	/* This method is called from ClientThread because the user logged out
   	   of an HTTP session.
	   Delete this User object and userid.  
	*/
	User usr = (User)users.get(id);
	if (usr!=null && usr.getTools().length==0)
	{
	    deleteUserid(usr.getNickname()); 
	    deleteUser(id);
	    usr = null;	
	}
	// Notify chat to terminate.
	sendMsg("logout", id);
    }

    // Called for force a ClientThread to shut down its connection.
    public synchronized void killConnection(String nick)
    {
	String key;
	User usr;
	ClientThread thr;
	for (Enumeration e=users.keys(); e.hasMoreElements(); )
	{
	    key = (String)e.nextElement();
	    usr = (User)users.get(key);
	    if (usr != null && usr.getNickname().equals(nick))
	    {
		for (Enumeration enum=conns.keys(); enum.hasMoreElements(); )
		{
		    thr = (ClientThread)enum.nextElement();
		    if (conns.get(thr).equals(key))
		    { 	 
			thr.closeConnection();
			return;
		    }
		}	
	    }
	} 
    }

    public synchronized String invalidateOldSession(String thisId,
	String nick, String name, String senderHost)
    {
	String key;
	User usr;

	/* Get the keys (sessionId) in the 'users' hashtable.  If the key is
	   not thisId, the associated User object is from a different session.
	   In this case, if the nickname is the same as 'nick' and the real
	   names match, log this user out.
	*/  
	for (Enumeration e=users.keys(); e.hasMoreElements(); )
	{
	    key = (String)e.nextElement();
	    usr = (User)users.get(key);
	    if (usr != null && !key.equals(thisId))
	    {
		String nickname = usr.getNickname(); 
		if (nickname.startsWith(nick) && usr.getRealname().equals(name))
	        {
		    /* Send chat applet at usr's senderHost a msg to quit IRC,
	               invalidate its session with PCCEServlet, and terminate.
		    */
		    ClientThread thr = (ClientThread)conns.get(key);
		    if (thr != null)	
	            {
			if (senderHost.equals(usr.getHost()))
		           thr.sendMsg("invalidate local"); 
			else
			   thr.sendMsg("invalidate remote");
		    }
	                
		    // Delete the User info from our internal tables.
	            deleteUserid(nickname);
	            deleteUser(key);
		    usr = null;
		    return nickname;
		}
	    } 
	} 
	return null;
    } 

    public synchronized void changeLoginInfo(String id, String nick, String pass)
    {
	User usr = (User)users.get(id);
	if (usr != null)
	{
	   usr.setNickname(nick);
	   usr.setPasswd(pass);
	   // Tell the chat applet to send a NICK msg to ircd.	
	   sendMsg("nick " + nick, id);
	}
    }

    public synchronized String getUserInfo(String id)
    {
	User usr = (User)users.get(id);
	if (usr != null)
	   return(usr.getUserInfo()); 
	return null;
    }

    public synchronized String getUserInfo(ClientThread thr)
    {
	String sessionId = (String)conns.get(thr);
	if (sessionId == null)
	    return null;
	return getUserInfo(sessionId);
    } 

    // Return all User objs for a nick that starts with or equals the given nick. 
    public synchronized Vector getAllUsers(String nick)
    {
	Vector v = new Vector();
	for (Enumeration e=users.elements(); e.hasMoreElements();)
	{
	    User usr = (User)e.nextElement();
	    if (usr.getNickname().startsWith(nick))
	        v.add(usr);	
	}
	return v;
    } 	 

    // Return a User object for the given nick.
    public synchronized User getUser(String nick)
    {
	for (Enumeration e=users.elements(); e.hasMoreElements();)
	{
	    User usr = (User)e.nextElement();
	    if (usr.getNickname().equals(nick))
	        return usr;
	}
	return null;
    } 

    // Return a User object for the given ClientThread, an open connection.
    public synchronized User getUser(ClientThread thr)
    {
	String sessionId = (String)conns.get(thr);
	if (sessionId == null)
	    return null;
	return((User)users.get(sessionId)); 
    }

    public synchronized String getOnlineUsers(String tool)
    {
	StringBuffer buff = new StringBuffer("online");
	String key;
	User usr;

	if (users.isEmpty())
	{
	    buff.append(" NONE");
	    return(buff.toString()); 
	}
	buff.append(" SUCCESS ");  
	for (Enumeration e=users.keys(); e.hasMoreElements();)
	{
	    key = (String)e.nextElement();
	    usr = (User)users.get(key);
	    if (usr!=null && usr.isRunning(tool))
		buff.append(usr.getUserInfo() + "|");
	}
	// strip off final "|"
	String reply = buff.toString();
	if (reply.endsWith("|"))
	    reply = reply.substring(0, reply.length()-1);
	return reply;
    }

    public synchronized String getToolUsers(String tool)
    {
	StringBuffer buff = new StringBuffer();
	String key;
	User usr;

	if (users.isEmpty())
	{
	    buff.append("NONE");
	    return(buff.toString()); 
	}
	for (Enumeration e=users.keys(); e.hasMoreElements();)
	{
	    key = (String)e.nextElement();
	    usr = (User)users.get(key);
	    if (usr!=null && usr.isRunning(tool))
	    {
		buff.append(tool+"user " + usr.getNickname() + " " + 
		    usr.getAvail() + ":" + usr.getLocation() + " ");
		if (usr.authenticated())
		    buff.append("true [");
		else
		    buff.append("false [");
		buff.append(usr.getRealname() + "] [" + usr.getEmail() + "] [" +
		    usr.getCompany() + "] [" + usr.getTitle() + "] [" +
		    usr.getGroupsAsString() + "] [" + usr.getDescription() + "]|");
	    }
	}
	// strip off final "|"
	String reply = buff.toString();
	if (reply.endsWith("|"))
	    reply = reply.substring(0, reply.length()-1);
	return reply;
    }

    public synchronized boolean sendMsg(String msg, String sessionId) 
    {
	/* Forward msg to a client with given 'sessionId'.  Use sessionId to look 
	   up the connection in a hashtable. Then send msg to that connection. 
  	*/  
	ClientThread thr;
	String id;
	for (Enumeration e=conns.keys(); e.hasMoreElements(); )
	{
	    thr = (ClientThread)e.nextElement();
	    id = (String)conns.get(thr);
	    if (id.equals(sessionId))
	    {
	        thr.sendMsg(msg);
	        return true;	
	    }
	}
	return false;
    }	

    // Send a msg to everyone connected.
    public void sendAll(String msg)
    {
	ClientThread thr;
	synchronized(conns) {
	    for (Enumeration e=conns.keys(); e.hasMoreElements(); )
	    {
	        thr = (ClientThread)e.nextElement();
	        thr.sendMsg(msg);
	    }		
	}
    } 

    public void addSessionId(ClientThread th, String id)
    {
	synchronized(conns) {
	    conns.remove(th);
	    conns.put(th, id);
	}
    }

    // Put a note for a user who is offline or online.
    public synchronized String putNote(String sender, String msg)
    {
	String target;
	int i = msg.indexOf(" ");
	if (i < 0)
	    return("note BADREQUEST"); 
	target = msg.substring(0,i);
	msg = msg.substring(i+1);
	i = target.indexOf("_");
	if (i > -1)
	    target = target.substring(0, i); 
	Vector usrs = getAllUsers(target);
	if (usrs != null && !usrs.isEmpty())
	{
	    for (i=0; i<usrs.size(); i++)
	    {
	        // Send note to the targeted usr
		User usr = (User)usrs.elementAt(i);
	        sendMsg("notes SUCCESS " + sender + " " + msg, usr.getSessionId());
	    }
	} 
	// Write the note to 'target.msg' file.
	FileHandler fh2 = new FileHandler(target + ".msg");
	try {
	    fh2.writeToFile("note " + sender + " " + msg);
	} catch(Exception e) {
	    return("putnote FAILURE " + sender);	 
	}
	return("putnote SUCCESS " + sender);
    } 

    // Retrieve and return all the notes left for a user.
    public synchronized String getNotes(User usr)
    {
	StringBuffer buff = new StringBuffer("notes SUCCESS ");
	if (usr != null)
	{
	   String notes = usr.getNotes();
	   if (notes != null)
	       buff.append(notes);
	}
	String reply = buff.toString();
	if (reply.endsWith("|"))
	{
	   int i = reply.lastIndexOf("|");
	   reply = reply.substring(0, i);
	}
	return reply;
    }

    // Set the user availability and notify all the users.
    public synchronized void setUserStatus(String id, String tool, String status,
	ClientThread thrId)
    {
	User usr = (User)users.get(id);
	if (usr != null)
	{
	    if (usr.isRunning(tool))
	    {
		usr.setAvail(status);
		ClientThread thr; 
		for (Enumeration e=conns.keys(); e.hasMoreElements(); )
		{
		    thr = (ClientThread)e.nextElement();
		    if (thr != null && thr != thrId)
			thr.sendMsg("status SUCCESS " + usr.getNickname() + " " + status);
		}
	    }
	}
    } 

    // Set the user's location and notify all the users.
    public synchronized void setUserLocation(String id, String tool, String loc,
	ClientThread thrId)
    {
	User usr = (User)users.get(id);
	if (usr != null)
	{
	    if (usr.isRunning(tool))
	    {
		usr.setLocation(loc);
		ClientThread thr; 
		for (Enumeration e=conns.keys(); e.hasMoreElements(); )
		{
		    thr = (ClientThread)e.nextElement();
		    if (thr != null && thr != thrId)
			thr.sendMsg("location SUCCESS " + usr.getNickname() + 
			    " " + loc);
		}
	    }
	}
    } 

    // Set the channel to 'open' or 'closed' and notify all the users.
    public synchronized void setDoor(String id, String name, String tool, 
        boolean closed, ClientThread thrId)
    {
	Channel ch = (Channel)channels.get(name);
	if (ch != null)
	    ch.setClosed(closed);
	
	User usr = (User)users.get(id);
	if (usr != null)
	{
	    if (usr.isRunning(tool))
	    {
		ClientThread thr; 
		for (Enumeration e=conns.keys(); e.hasMoreElements(); )
		{
		    thr = (ClientThread)e.nextElement();
		    if (thr != thrId)
			thr.sendMsg("door SUCCESS " + name + " " + closed);
		}
	    }
	}
    } 

    public synchronized void deleteConnection(ClientThread th)
    {
	/* Get the id associated w/ this thread. Use this id as the key into the 
	   'users' hashtable to get the User object.  Then remove the tool 
	   associated w/ this thread from the User object and delete 'th' from the
           hashtable of connections.  Notify all the other users.
	*/
	 	   
	String tool = null;
	String nick = null;
	String id = (String)conns.get(th);
	if (id != null)
	{
	    User usr = (User)users.get(id);
	    if (usr != null)
	    {
		tool = th.getTool();
		nick = usr.getNickname();
		usr.removeTool(tool);	   
		if (usr.getTools().length == 0)
		{
		    deleteUserid(nick);
		    deleteUser(id);
		}
		if (tool != null && nick != null)
		{
	            for (Enumeration e=conns.keys(); e.hasMoreElements(); )
	            {
			ClientThread thr = (ClientThread)e.nextElement();
			if (thr != th)
			    thr.sendMsg("userquit SUCCESS " + tool + " " + nick);
	            } 	
		}
	    }
	}
	conns.remove(th);
    }

    // Methods to handle chat channels. 
    public synchronized String getChannels(String type)
    {
	StringBuffer buff  = new StringBuffer(type + " SUCCESS ");
	String reply;
	String key;
	Channel ch;
	for (Enumeration e=channels.keys(); e.hasMoreElements();)
	{
	    key = (String)e.nextElement();
	    ch = (Channel)channels.get(key);	
	    if (type.equals("pchan") && !ch.isPermanent())
		continue;
	    else if (type.equals("tchan") && ch.isPermanent())
		continue;
	    buff.append(ch.getName() + " " + ch.getCreator() + " (" + ch.getDescription() +
		")|");	
	} 
	reply = buff.toString();
	int endIndx = reply.lastIndexOf("|");
	if (endIndx > -1)
	    reply = reply.substring(0, endIndx);
	return reply;
    }

    // Return the state of all known channels.
    public synchronized String getChannels()
    {
	if (channels == null || channels.isEmpty())
	    return null;
	StringBuffer buff = new StringBuffer();
	String key;
	Channel ch;
	for (Enumeration e=channels.keys(); e.hasMoreElements();)
	{
	    buff.append("venue ");	
	    key = (String)e.nextElement();
	    ch = (Channel)channels.get(key);
	    buff.append(ch.getName() + " ");
	    if (ch.isPermanent())
		buff.append("P ");
	    else
		buff.append("T ");
	    if (ch.isClosed())
		buff.append("C ");
	    else
		buff.append("O ");
	    buff.append(ch.getCreator() + " " + "[" + ch.getDescription() +
		"] [" + ch.getTopic() + "]|");
	}
	String reply = buff.toString();
	return(reply.substring(0, reply.length()-1));
    }

    // Needed for InterGroup
    public synchronized String getMulticastAddrs()
    {
	StringBuffer buff = new StringBuffer("mcastaddrs SUCCESS ");
	String reply;
	String key;
	Channel ch;
	for (Enumeration e=channels.keys(); e.hasMoreElements();)
	{
	    key = (String)e.nextElement();
	    ch = (Channel)channels.get(key);
	    buff.append(ch.getName() + " ");
	    buff.append(ch.getMcastAddress() + " ");
	    buff.append(ch.getMcastPort() + "|");
	} 
	reply = buff.toString();
	int endIndx = reply.lastIndexOf("|");
	if (endIndx > -1)
	    reply = reply.substring(0, endIndx);
	return reply;
    }

    public synchronized void addChannel(String id, String tool, ClientThread thrId,
	String nm, String desc, String creator, boolean perm, boolean pvt)
    {
	channels.put(nm, new Channel(nm,desc,creator,perm,pvt)); 

	// Notify all users running the tool (chat).
	User usr = (User)users.get(id);
	if (usr != null)
	{
	    if (usr.isRunning(tool))
	    {
		ClientThread thr; 
		for (Enumeration e=conns.keys(); e.hasMoreElements(); )
		{
		    thr = (ClientThread)e.nextElement();
		    if (thr != thrId)
			thr.sendMsg("newvenue " + nm + " " + perm + " " + pvt);
		}
	    }
	}
    }

    public synchronized void deleteChannel(String name)
    {
	Channel ch = (Channel)channels.get(name);
	if (ch != null)
	{
	    channels.remove(name);
	    ch = null;
	} 
    }

    protected ClientThread connectToClient() 
    {
	ClientThread conn = null;
        SSLSocket sock;

	// Accept client connection request.
	Vector DN = comm.doSSLAccept();
	sock = comm.getClientSSLSock();
	if (sock == null)
	    return null;

  	// Create a thread to handle this connection.
	try {
	    conn = new ClientThread(sock, this, fh, keyFile, casFile, password, DN);
	    conns.put(conn, "0");
	    conn.start();
	} catch(Exception e) {
	    System.err.println("Couldn't create ClientThread");
	    e.printStackTrace();
	    return null;
	}
	return conn;
    }	 
}
