// This class encapsulates information for a user (a client).

import pccejava.*;
import java.util.*;

public class User {
    /* Since Http sessionId's are used by web servers to keep session state, 
       use the sessionId to uniquely identify a client (each instance of a 
       browser has a unique session id for the HttpSession.)
    */
    private String sessionId = null;
	
    // nickname currently used with IRC 
    private String nickname = null;	
    // the tools that this user is running	
    private Vector tools = new Vector();
    private String realname = null;
    private String company = null;
    private String title = null;
    private String email = null;
    private String passwd = null;
    private String description = "NONE"; 

    // host from which the user logged in
    private String host = null;	
   
    // The set of collaboration groups the user is a member of.
    Set groups = new TreeSet();

    // Availability: available, busy, away...
    private String avail = "available";

    // Online location: home, work, mobile 
    private String location = "Work";

    // Indicate if user presented an X.509 certificate. 
    private boolean authenticated = false;

    // Notes left for this user.
    Vector notes = new Vector(10);

    public User(String id) { sessionId = id; }
	
    public void setUserInfo(String n, String p, String r, String e, String c,
	String t, String h, String app)
    {
	nickname = n;
	realname = r;
	company = c;
	title = t;
	email = e;
	passwd = p;
	host = h;
	tools.add(app);
    }

    public void setAuthenticated(boolean auth) { authenticated = auth; }  

    public boolean authenticated() { return authenticated; }

    public String getSessionId() { return sessionId; }

    public String getUserInfo() 
    { 
	String auth = "false";
	if (authenticated)
	    auth = "true";
	return(nickname + " " + avail + " " + auth +
	   " [" + realname + "] [" + email + "] [" + company + "] [" + 
	   title + "]");
    }

    public String getAllInfo()
    {
	String auth = "false";
	if (authenticated)
	    auth = "true";
	return(nickname + " " + avail + ":" + location + " " + auth + " [" +
	    realname + "] [" + email + "] [" + company + "] [" + title + "] [" +
	    getGroupsAsString() + "]"); 
    }
  
    // Return the vector of tools the user is running.
    public String[] getTools()
    {
	String[] apps = new String[tools.size()];
	Enumeration e = tools.elements();
	for (int i=0; i<tools.size(); i++)
	    apps[i] = (String)e.nextElement();
	return apps;
    }

    // Returns msgs from the user's .msg file and those in memory. 
    public String getNotes()
    {
	String nick = nickname;
	int i = nick.indexOf("_");
	if (i > -1)
	    nick = nick.substring(0, i);
	FileHandler fh = new FileHandler(nick + ".msg");
	return(fh.findNotes());
    }

    // Remove an entry from this user's notes (.msg) file.
    public void deleteNote(String note)
    {
	String nick = nickname;
	int i = nick.indexOf("_");
	if (i > -1)
	    nick = nick.substring(0, i);
	FileHandler fh = new FileHandler(nick + ".msg");
	try {
	    fh.delete(note);
	} catch(Exception e) { 
	    System.err.println("User::deleteNote() Exception: " + 
		e.getMessage());
	}
    }	

    // Return true if the given tool is already running, else return false.
    public boolean isRunning(String tool) { return tools.contains(tool); }	

    public void addTool(String t) { tools.add(t); }

    public void removeTool(String t) { tools.removeElement(t); }

    public void addGroup(String name) { groups.add(name); }

    public void deleteGroup(String name) { groups.remove(name); }
    public void clearGroups() { groups.clear(); }

    public String getGroupsAsString() 
    {
	if (groups.isEmpty())
	    return "NONE";
	StringBuffer grps = new StringBuffer(); 
	for (Iterator i=groups.iterator(); i.hasNext(); )
	    grps.append(i.next().toString() + " ");		
	String reply = grps.toString().trim();
	if (reply.endsWith("]"))
	    reply = reply.substring(0, reply.length()-1);
	return reply; 
    } 

    public void setPasswd(String pass) { passwd = pass; }
    public String getPasswd() { return passwd; }

    public void setRealname(String n) { realname = n; }
    public String getRealname() { return realname; }

    public void setNickname(String nick) { nickname = nick; } 
    public String getNickname() { return nickname; }

    public void setHost(String h) { host = h; }
    public String getHost() { return host; }	

    public void setAvail(String status) { avail = status; }
    public String getAvail() { return avail; } 

    public void setLocation(String loc) { location = loc; }
    public String getLocation() { return location;} 

    public void setEmail(String addr) { email = addr; }
    public String getEmail() { return email;} 

    public void setTitle(String t) { title = t; }
    public String getTitle() { return title;} 

    public void setCompany(String org) { company = org; }
    public String getCompany() { return company;} 

    public void setDescription(String desc) { description = desc; }
    public String getDescription() { return description;} 
}
